/*
 * pycups - Python bindings for CUPS
 * Copyright (C) 2002  Tim Waugh <twaugh@redhat.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <Python.h>
#include <cups/cups.h>
#include <cups/language.h>

#include <locale.h>

static PyObject *
construct_job_list (void)
{
	PyObject *result = PyList_New (0);
	http_encryption_t encryption = cupsEncryption ();
	http_t *http = httpConnectEncrypt (cupsServer (), ippPort (),
					   encryption);
	ipp_t *request = ippNew(), *answer;
	ipp_attribute_t *attr;
	const char *attributes[] = {
		  "job-id",
		  "job-k-octets",
		  "job-name",
		  "time-at-creation",
		  "job-printer-uri",
		  "job-originating-user-name",
	};
	char *lang = setlocale (LC_MESSAGES, NULL);

	if (!http)
		return NULL;

	ippAddString (request, IPP_TAG_OPERATION, IPP_TAG_CHARSET,
		      "attributes-charset", NULL, "utf-8");
	ippAddString (request, IPP_TAG_OPERATION, IPP_TAG_LANGUAGE,
		      "attributes-natural-language", NULL, lang);
	ippAddStrings (request, IPP_TAG_OPERATION, IPP_TAG_KEYWORD,
		       "requested-attributes",
		       sizeof (attributes) / sizeof (attributes[0]),
		       NULL, attributes);
	ippAddString (request, IPP_TAG_OPERATION, IPP_TAG_URI, "job-uri",
		      NULL, "ipp://localhost/jobs/");

	request->request.op.operation_id = IPP_GET_JOBS;
	request->request.op.request_id = 1;
	answer = cupsDoRequest (http, request, "/");
	httpClose(http);
	if (!answer || answer->request.status.status_code > IPP_OK_CONFLICT)
		return NULL;

	for (attr = answer->attrs; attr; attr = attr->next) {
		PyObject *dict;
		int got_job = 0;

		while (attr && attr->group_tag != IPP_TAG_JOB)
			attr = attr->next;

		if (!attr)
			break;

		dict = PyDict_New ();
		for (; attr && attr->group_tag == IPP_TAG_JOB;
		     attr = attr->next) {
			PyObject *val = NULL;

			if ((!strcmp (attr->name, "job-id") ||
			     !strcmp (attr->name, "job-k-octets") ||
			     !strcmp (attr->name, "time-at-creation")) &&
			    attr->value_tag == IPP_TAG_INTEGER) {
				int integer = attr->values[0].integer;
				val = PyInt_FromLong (integer);
				got_job = 1;
			}
			else if ((!strcmp (attr->name,
					   "job-originating-user-name") ||
				  !strcmp (attr->name, "job-name")) &&
				 attr->value_tag == IPP_TAG_NAME) {
				char *s = attr->values[0].string.text;
				val = PyString_FromString (s);
			}

			if (val) {
				PyDict_SetItemString (dict, attr->name, val);
				Py_DECREF (val);
			}
		}

		if (got_job) {
			PyList_Append (result, dict);
		} else
			Py_DECREF (dict);

		if (!attr)
			break;
	}

	return result;
}

static PyObject *
construct_queue_dict (void)
{
	PyObject *result = PyDict_New ();
	http_encryption_t encryption = cupsEncryption ();
	http_t *http = httpConnectEncrypt (cupsServer (), ippPort (),
					   encryption);
	ipp_t *request = ippNew(), *answer;
	ipp_attribute_t *attr;
	const char *attributes[] = {
		"printer-name",
		"printer-type",
		"printer-location",
		"printer-info",
		"printer-make-and-model",
		"printer-state",
		"printer-uri",
	};
	char *lang = setlocale (LC_MESSAGES, NULL);

	if (!http)
		return NULL;

	ippAddString (request, IPP_TAG_OPERATION, IPP_TAG_CHARSET,
		      "attributes-charset", NULL, "utf-8");
	ippAddString (request, IPP_TAG_OPERATION, IPP_TAG_LANGUAGE,
		      "attributes-natural-language", NULL, lang);
	ippAddStrings (request, IPP_TAG_OPERATION, IPP_TAG_KEYWORD,
		       "requested-attributes",
		       sizeof (attributes) / sizeof (attributes[0]),
		       NULL, attributes);
	request->request.op.operation_id = CUPS_GET_PRINTERS;
	request->request.op.request_id = 1;
	answer = cupsDoRequest (http, request, "/");
	httpClose(http);
	if (!answer || answer->request.status.status_code > IPP_OK_CONFLICT)
		return NULL;

	for (attr = answer->attrs; attr; attr = attr->next) {
		PyObject *dict;
		char *printer = NULL;

		while (attr && attr->group_tag != IPP_TAG_PRINTER)
			attr = attr->next;

		if (!attr)
			break;

		dict = PyDict_New ();
		for (; attr && attr->group_tag == IPP_TAG_PRINTER;
		     attr = attr->next) {
			PyObject *val = NULL;

			if (!strcmp (attr->name, "printer-name") &&
			    attr->value_tag == IPP_TAG_NAME)
				printer = attr->values[0].string.text;
			else if ((!strcmp (attr->name, "printer-type") ||
				  !strcmp (attr->name, "printer-state")) &&
				 attr->value_tag == IPP_TAG_ENUM) {
				int ptype = attr->values[0].integer;
				val = PyInt_FromLong (ptype);
			}
			else if ((!strcmp (attr->name,
					   "printer-make-and-model") ||
				  !strcmp (attr->name, "printer-info") ||
				  !strcmp (attr->name, "printer-location") ||
				  !strcmp (attr->name,
					   "printer-state-message")) &&
				 attr->value_tag == IPP_TAG_TEXT) {
				val = PyString_FromString (attr->values[0].string.text);
			}
			else if (!strcmp (attr->name,
					  "printer-state-reasons") &&
				 attr->value_tag == IPP_TAG_KEYWORD) {
				val = PyString_FromString (attr->values[0].string.text);
			}
			else if (!strcmp (attr->name,
					  "printer-is-accepting-jobs") &&
				 attr->value_tag == IPP_TAG_BOOLEAN) {
				int b = attr->values[0].boolean;
				val = PyInt_FromLong (b);
			}
			else if ((!strcmp (attr->name,
					   "printer-up-time") ||
				  !strcmp (attr->name,
					   "queued-job-count")) &&
				 attr->value_tag == IPP_TAG_INTEGER) {
				int u = attr->values[0].integer;
				val = PyInt_FromLong (u);
			}

			if (val) {
				PyDict_SetItemString (dict, attr->name, val);
				Py_DECREF (val);
			}
		}

		if (printer) {
			PyDict_SetItemString (result, printer, dict);
		} else
			Py_DECREF (dict);

		if (!attr)
			break;
	}

	return result;
}

static int
do_model_compare (const char *a, const char *b)
{
  const char *digits = "0123456789";
  char quick_a, quick_b;
  while ((quick_a = *a) != '\0' && (quick_b = *b) != '\0') {
    int end_a = strspn (a, digits);
    int end_b = strspn (b, digits);
    int min;
    int a_is_digit = 1;
    int cmp;

    if (quick_a != quick_b && !isdigit (quick_a) && !isdigit (quick_b)) {
      if (quick_a < quick_b) return -1;
      else return 1;
    }

    if (!end_a) {
      end_a = strcspn (a, digits);
      a_is_digit = 0;
    }

    if (!end_b) {
      if (a_is_digit)
	return -1;
      end_b = strcspn (b, digits);
    } else if (!a_is_digit)
      return 1;

    if (a_is_digit) {
      int n_a = atoi (a), n_b = atoi (b);
      if (n_a < n_b) cmp = -1;
      else if (n_a == n_b) cmp = 0;
      else cmp = 1;
    } else {
      min = end_a < end_b ? end_a : end_b;
      cmp = strncmp (a, b, min);
    }

    if (!cmp) {
      if (end_a != end_b)
	return end_a < end_b ? -1 : 1;

      a += end_a;
      b += end_b;
      continue;
    }

    return cmp;
  }

  return 0;
}

static PyObject *
model_sort (PyObject *self, PyObject *args)
{
  PyObject *ret;
  char *a, *b;

  if (!PyArg_ParseTuple (args, "ss", &a, &b))
    return NULL;

  return Py_BuildValue ("i", do_model_compare (a, b));
}

static PyObject *
pycups_get_queues (PyObject *self, PyObject *args)
{
	PyObject *ret;
	if (!PyArg_ParseTuple (args, ""))
		return NULL;

	ret = construct_queue_dict ();
	if (!ret) {
		Py_INCREF (Py_None);
		ret = Py_None;
	}
	return ret;
}

static PyObject *
pycups_get_jobs (PyObject *self, PyObject *args)
{
	PyObject *ret;
	if (!PyArg_ParseTuple (args, ""))
	    return NULL;

	ret = construct_job_list ();
	if (!ret) {
		Py_INCREF (Py_None);
		ret = Py_None;
	}
	return ret;
}

static PyMethodDef PycupsMethods[] = {
	{"get_queues", pycups_get_queues, METH_VARARGS,
	 "No arguments.\n"
	 "Returns a dict, indexed by name, of dicts representing browsed\n"
	 "queues, indexed by attribute."},
	{"get_jobs", pycups_get_jobs, METH_VARARGS,
	 "No arguments.\n"
	 "Returns a list of dicts representing jobs."},
	{"model_sort", model_sort, METH_VARARGS, ""},
	{NULL, NULL, 0, NULL}
};

void
initpycups (void)
{
	PyObject *m = Py_InitModule ("pycups", PycupsMethods);
	PyObject *d = PyModule_GetDict (m);
	PyObject *c;

	c = PyInt_FromLong (CUPS_PRINTER_REMOTE);
	PyDict_SetItemString (d, "CUPS_PRINTER_REMOTE", c);
	Py_DECREF (c);

	c = PyInt_FromLong (CUPS_PRINTER_DEFAULT);
	PyDict_SetItemString (d, "CUPS_PRINTER_DEFAULT", c);
	Py_DECREF (c);
}
